/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1999 University of California, Santa Barbara           */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

# define _REENTRANT
# include <stdio.h>
# include <malloc.h>
# include <sys/types.h>
# include <sys/ipc.h>
# include <sys/shm.h>
# include <ulocks.h>
# include <sys/time.h>
# include <sys/sysmp.h>
# include <sys/prctl.h>

# include "ukey.h"

static char **shmfile = NULL;
static int *shmid;
static void **shmvaddr;
static void **ap;
extern int thrMPI_Nproc;
extern usptr_t *thrMPI_sc_handle;

void thrMPI_exit(int);

void anl_shmem_ginit(void)
{
	shmfile=(char **)malloc(sizeof(char *)*thrMPI_Nproc);
	shmid=(int *)malloc(sizeof(int)*thrMPI_Nproc);
	shmvaddr=(void **)malloc(sizeof(void *)*thrMPI_Nproc);
	ap=(void **)malloc(sizeof(void *)*thrMPI_Nproc);

  	if( usconfig(CONF_INITSIZE, 2097152) == -1 ){
        	perror("usconfig(CONF_INITSIZE)");
        	thrMPI_exit(-1);
  	}

  	if( usconfig(CONF_INITUSERS, thrMPI_Nproc) == -1 ){
        	perror("usconfig(CONF_INITUSERS,nProc)");
        	thrMPI_exit(-1);
  	}

  	usconfig(CONF_LOCKTYPE, US_NODEBUG); 

	/* the following actually should be counted as bar_ginit, */
	/* but keep with shmem_ginit() for convenience */
  	thrMPI_sc_handle = usinit (mktemp("/usr/tmp/lckXXXXXX"));
  	if (thrMPI_sc_handle == NULL) {
    		perror ("usinit");
    		thrMPI_exit (-1);
  	}
}

void anl_shmem_uinit(int myid, size_t shmsize)
{
	key_t shmkey;
	usptr_t *usptr;
	char *template;

	template=(char *)malloc(16);
	strcpy(template, "/tmp/shmXXXXXX");

	shmkey = (key_t)getpid();
	shmkey = (shmkey<<16)|shmkey;

	if((shmid[myid]=shmget(shmkey, shmsize, IPC_CREAT|SHM_R|SHM_W|044)) == -1){
		perror("shmget");
		exit(1);
	}
	printf("shmem_init: shmid[%d]=%d\n", myid, shmid[myid]); 
	/*
	 * The segment needs to be attached only once; all
	 * attached shared memory segments are inherited by 
	 * forked off processes.
	 */

	if((shmvaddr[myid]=shmat(shmid[myid], (void *)0, 0)) == (void *)(-1)){
		perror("shmat");
		exit(1);
	}

	shmfile[myid] = mktemp(template);
	/** printf("filename=%s", shmfile[myid]); **/
	if((usptr=usinit(shmfile[myid])) == NULL){
		perror("usinit");
		exit(1);
	}

	if((ap[myid]=acreate(shmvaddr[myid], shmsize, MEM_SHARED, usptr, NULL)) == NULL){
		perror("acreate");
		exit(1);
	}
}

void *
anl_malloc(int myid, size_t size)
{
# if (_DEBUG)
	void *ret;
	ret=amalloc(size, ap[myid]);
	if (!ret) 
		fprintf(stderr, "out of shmem\n");
	return ret;
# else
	return( amalloc(size, ap[myid]) );
# endif
}

void
anl_free(int myid, void *ptr)
{
	afree(ptr, ap[myid]);
}

void
anl_shmem_uend( int myid)
{
	adelete(ap[myid]);
	shmctl(shmid[myid], IPC_RMID);
	unlink(shmfile[myid]);
}

void
anl_shmem_gend()
{
	free(ap);
	free(shmid);
	free(shmfile);
	free(shmvaddr);
}

