/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <stdio.h>
# include <time.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "global.h"
# include "mpitrans.h"
# include "symbol.h"

FILE *INS; /* *.ins */
FILE *INIT; /* thrUserInit.c */
FILE *MAIN; /* thrMain.c */
char prefix[9]="_G000000";
int randomize=0;


StrBuf *symnames;

char filename[MAX_FILENAME+1]=""; /* source file name, need to parse #file directive */
char insfname[MAX_FILENAME+1];
char userFile[100][MAX_FILENAME+1]; /* currently only one file */

int LEVEL=0;
STG_CLS last_stg_cls=NONE;
Bool look_for_tag=FALSE;
Bool parse_attrib_list=FALSE;
static char *ofilename=NULL;
static char *welcome="Thanks for using MPI Translator!\n\t--by Tang Hong, ";
static char *date="05/26/98 Version 1.0, All rights reserved.";
static char mainfilename[MAX_FILENAME+1];
static char initfilename[MAX_FILENAME+1];
static char id_suffix[101];

int yyparse();

void yerror(char *msg)
{
	fprintf(stderr, "%s\n", msg);
}

int cmdline_parse(argc,argv)
int argc;
char *argv[];
{
	int i=0;
	time_t tmpt;
	char 	*today;
	int j;

	if (argc==1)
		return 0;

	while (--argc) {
		if (argv[++i][0]=='-') {
			switch (argv[i][1]) {
			case 'h':
			case 'H': return 0; /* do_help */
			case 'r':
			case 'R': 
				if (!--argc) return 0;
				randomize=atoi(argv[++i]);
				break;
			case 'e':
			case 'E': 
				break;
			case 'i':
			case 'I': 
				if (!--argc) return 0;
				strncpy(id_suffix, argv[++i], 100);
				break;
			case 'o':
			case 'O': if (!--argc) return 0;
				ofilename=argv[++i];
			 	break;
			case 'f':
			case 'F': 
				j=0;
				do {
					if (!--argc) return 0;
					strncpy(userFile[j], argv[++i], MAX_FILENAME);
					j++;
				} while (argv[i+1] && argv[i+1][0]!='-');
			 	break;
			case '\0': 
			case '-' : 
				if (strlen(filename)) {
					yerror("source file specified!");
					return 0;
				}
				else
					strcpy(filename,"-");
				break;
			default: return 0;
			}
		}
		else {
			if ( strlen(filename) && (ofilename) ) {
				yerror("Source and destiny files have been specified already!");
				return 0;
			}
			if (strlen(filename))
				ofilename=argv[i];
			else
				strncpy(filename,argv[i],MAX_FILENAME);
		}
	}
	if (!strlen(filename)) {
		yerror("No source file specified!");
		return 0;
	}
	if (filename[0]=='-' && filename[1]=='\0')  {
		yyin=stdin;
		strcpy(filename,"$STDIN$");
	}
	else if ((yyin = fopen (filename,"r")) == NULL) {
		yerror("Cannot open source file!\n");
		return -1;
	}

	strcpy(mainfilename, filename);
	strcat(mainfilename, ".thrMain.c");
	strcpy(initfilename, filename);
	strcat(initfilename, ".thrUserInit.c");

	if ( (ofilename==NULL) || ((ofilename[0]=='-') && (ofilename[1]=='\0')) )
		yyout=stdout;
	else if ((yyout = fopen (ofilename,"w")) == NULL) {
		yerror("Cannot open destiny file!\n");
		return -1;
	}
	tmpt=time(NULL);
	today=ctime(&tmpt);
	today[strlen(today)-1]='\0';
	fprintf(yyout,"/* MPItrans processed. Date : %s */\n",today);
	fprintf(yyout,"int thrMPI_createkey(void);\n");
	fprintf(yyout,"void thrMPI_setval(int, void *);\n");
	fprintf(yyout,"void *thrMPI_getval(int);\n");

	if (strlen(userFile[0])==0)
		strcpy(userFile[0], filename);

	return 1;
}

void do_help()
{
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"\tmpitrans -[hH]: Print this screen.\n");
	fprintf(stderr,"\tmpitrans [options] source [destiny]\n");
	fprintf(stderr,"\tmpitrans [options] source [options]\n"); 
	fprintf(stderr,"[options] could be:\n");
	fprintf(stderr,"\t[-o destiny] : specify destiny file\n"); 
	fprintf(stderr,"\tBoth source and destiny are default to stdin/stdout. If file name is given as '-', stdin/stdout will be used.\n");


	return;
}
	
void init()
{
	char *program[]={
		"{",
		"	/* allocate memory for external keys */",
		NULL
	};
	char *program2[]={
		"{",
		"	int i;",
		"",
		"/* allocate static keys */",
		"	/* for (i=0; i<nSKEY; i++) */",
		"		/* thrMPI_createkey(); */",
		"# undef nSKEY",
		"",
		"/* allocate external keys */",
		NULL
	};
	int i;
	char *pc=filename;

	/* init prefix */
	if (randomize==0) {
		while (*pc) {
			randomize+=*pc;
			pc++;
		}
		srand((int)time(NULL)+randomize);
		randomize=rand()%1000000;
	}
	i=7;
	while (randomize) {
		prefix[i]='0'+randomize%10;
		randomize/=10;
		i--;
	}
	/* init INS */
	strcpy(insfname, filename);
	strcat(insfname, ".ins");
	INS=fopen(insfname,"w");
	fprintf(INS, "%s%s%s",
		"/* Generated by mpiTrans, for file ",
		filename,
		". This file is to be included in the main file. */\n");

	INIT=fopen(initfilename,"w");
	fprintf(INIT, "%s%s%s%s",
		"/* Generated by mpiTrans, for file ",
		filename,
		". */\n",
		"/* Do NOT compile! This file is not an independent C file. */\n");
	i=0;
	fprintf(INIT, "void thrUserInit%s(void)\n", id_suffix);
	while (program[i])  {
		fprintf(INIT, "%s\n", program[i]);
		i++;
	}

	MAIN=fopen(mainfilename,"w");
	i=0;
	fprintf(MAIN, "void thrMainInit%s(void)\n", id_suffix);
	while (program2[i])  {
		fprintf(MAIN, "%s\n", program2[i]);
		i++;
	}
	/**
	gvnames=new_strbuf();
	lvnames=new_strbuf();
	**/
	symnames=new_strbuf();
	symtab_init();

	return;
}

void cleanup()
{
	fprintf(yyout,"\n# include \"%s\"\n",initfilename);
	fprintf(yyout,"\n# include \"%s\"\n", mainfilename);
	fprintf(INIT,"}\n");
	fprintf(MAIN,"}\n");
	fprintf(INS, "#undef %s\n", prefix);
}

int main(argc,argv)
int argc;
char *argv[];
{
# if YYDEBUG
	extern yydebug;

	yydebug = 1;
# endif
	printf("%s%s\n",welcome,date);

	switch (cmdline_parse(argc,argv)) {
	case -1 : exit(-1);
	case 0  : do_help(); return 0;
	default :
		init();
		if (yyparse() || errmsgno) {
			cleanup(); 
			exit(-1);
		}
		cleanup(); 
		return 0;
	}
}
