/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include "comm.h"
# include "bcast.h"
# include "ukey.h"
# include "struct.h"
# include "mpi.h"

#ifdef MPISIM
#include "sim.h"
#endif

int thrMPI_Bcast(
	void *buffer,
	int count,
	MPI_Datatype datatype,
	int root,
	MPI_Comm comm)
{
	int myid=ukey_getid();
	int buflen=count*thrMPI_Usize(datatype);
	int ret;
#ifdef MPISIM
	int lseq=sim_comm_start(myid, Event_Bcast, comm, root, buflen);
#endif
	thrMPI_Comm *tcomm=comm_verify(comm, myid);

	if (tcomm==NULL) return MPI_ERR_COMM;

	ret=thrMPI_bco_bcast(tcomm->bc_ptr, myid, root, (char *)buffer, buflen
#ifdef MPISIM
			     , lseq
#endif
			     );
#ifdef MPISIM
	sim_comm_end(myid);
#endif
	switch(ret) {
		case 0  : return MPI_SUCCESS;
		case -1 : return MPI_ERR_INTERN;
		case -2 : 
# if (_DEBUG_)
			fprintf(stderr, "buflen=%d\n", buflen);
# endif
			return MPI_ERR_BUFFER;
		default : return MPI_ERR_INTERN;
	}
}
