/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include "comm.h"
# include "reduce.h"
# include "ukey.h"
# include "struct.h"
# include "mpi.h"

int thrMPI_Reduce(
	void *sendbuf,
	void *recvbuf,
	int count,
	MPI_Datatype datatype,
	MPI_Op op,
	int root,
	MPI_Comm comm)
{
	int myid=ukey_getid();
	thrMPI_Comm *tcomm=comm_verify(comm, myid);

	if (tcomm==NULL) return MPI_ERR_COMM;

	switch(thrMPI_rdco_reduce(tcomm->rdc_ptr, myid, root, sendbuf, recvbuf, count, datatype, op)) {
		case 0  : return MPI_SUCCESS;
		case -1 : return MPI_ERR_INTERN;
		case -2 : return MPI_ERR_BUFFER;
		default : return MPI_ERR_INTERN;
	}
}

int thrMPI_Allreduce(
        void *sendbuf,
        void *recvbuf,
        int count,
        MPI_Datatype datatype,
        MPI_Op op,
        MPI_Comm comm)
{
        int myid=ukey_getid();
        thrMPI_Comm *tcomm=comm_verify(comm, myid);

        if (tcomm==NULL) return MPI_ERR_COMM;

        switch(thrMPI_ardco_allreduce(tcomm->ardc_ptr, myid, sendbuf, recvbuf, count, datatype, op)) {
                case 0  : return MPI_SUCCESS;
                case -1 : return MPI_ERR_INTERN;
                case -2 : return MPI_ERR_BUFFER;
                default : return MPI_ERR_INTERN;
        }
}

