/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <stdio.h>
# include <stdlib.h>
#include <sys/types.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h> 
# include <ulocks.h>

# define USE_SPROC
# include "anl_shmem.h"
# include "ukey.h"
# include "twait.h"
# include "bcast.h"
# define PNO 16
# define NT 100000
/* # define NT 800 */
# define MSG_LEN 100

thrMPITCB **thrMPI_tcbs;
thrMPI_BcastObj *bco;
int thrMPI_Nproc=4; /* default */
int ready=0;
double *times;

double gettime();

extern int sigs, busy, prob;

void testbcast(void)
{
	char buf[MSG_LEN+1];
	int i,j;
	double t1, t2;

	int myid=ukey_getid();

	if (myid!=thrMPI_Nproc-1)
		anl_shmem_uinit(myid, 0x400000);

	thrMPI_tcbs[myid]=(thrMPITCB *)anl_malloc(sizeof(thrMPITCB));
	thrMPI_tcbs[myid]->intr=0;
	thrMPI_tcbs[myid]->status=thrMPI_IDLE;
	thrMPI_tcbs[myid]->pid=getpid();

	/* sysmp(MP_MUSTRUN_PID, PNO+myid, getpid()); */
	for (i=0; i<MSG_LEN; i++)
		buf[i]='a'+myid;

	buf[MSG_LEN]='\0';
	
	__add_and_fetch(&ready, 1);
	waitFlagEQ(&ready, thrMPI_Nproc);

	t1=gettime();
	for (i=0;i<NT;i++) {
		if (thrMPI_bco_bcast(bco, myid, i%thrMPI_Nproc, buf, MSG_LEN))
				printf("failed\n");
		/** printf("buf=%c...\n", buf[0]); **/
	}

	t2=gettime();

	times[myid]=t2-t1;

	if (myid!=thrMPI_Nproc-1)
		anl_shmem_uend();
	return;
}

main(int argc, char *argv[])
{
	int i;
	double t=0.0;

	if (argc<2) {
		printf("usage: %s #proc\n", argv[0]);
		exit(-1);
	}
	
	thrMPI_Nproc=atoi(argv[1]);

        if( usconfig(CONF_INITUSERS, thrMPI_Nproc) == -1 ){
                perror("usconfig(CONF_INITUSERS,%d)");
                exit(-1);
        }

	anl_shmem_ginit();
	ukey_setid(thrMPI_Nproc-1);
	anl_shmem_uinit(thrMPI_Nproc-1, 0x400000);
	bco=thrMPI_bco_create(thrMPI_Nproc);
	times=(double *)anl_malloc(sizeof(double)*thrMPI_Nproc);
	thrMPI_tcbs=(thrMPITCB **)anl_malloc(sizeof(thrMPITCB *)*thrMPI_Nproc);

	for (i=0; i<thrMPI_Nproc-1; i++) {
		ukey_thrcreate(testbcast, i);
	}

	testbcast();

	/* waitFlagEQ(&ready, thrMPI_Nproc); */

	for (i=0; i<thrMPI_Nproc-1; i++) {
		ukey_thrwait(thrMPI_tcbs[i]->pid);
		t+=times[i];
	}

	printf("thrMPI_Nproc=%d, msg len=%d, skewed bcast cost=%e\n", thrMPI_Nproc, MSG_LEN, t/thrMPI_Nproc/NT);
	printf("busy=%d, sigs=%d, prob=%d\n", busy, sigs, prob);
	anl_shmem_uend();
	anl_shmem_gend();
}
