/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT
# include <errno.h>
# include "twait.h"

typedef struct {
        int maxcnt;           /* maximum number of runners    */
	int c1;
	int c2;
} mytp_thrMPI_Bar;

/* mytp_bar_init -  init the barrier */
int mytp_bar_init(mytp_thrMPI_Bar*bp, int count) {
  int n, i;

  if (count<1) return(EINVAL);

  bp->maxcnt = count;
  bp->c1=bp->c2=0;

  return 0;
}

/* mytp_bar_wait - wait at a barrier for everyone to arrive */
/* not a safe implementation, but should not have wrap around */
/* for normal run to complete jobs. (Not good for OS) */
int mytp_bar_wait(mytp_thrMPI_Bar *bp) {
  /* check if I am the last? */
  __add_and_fetch(&(bp->c1), 1);
  waitFlagMOD(&(bp->c1), bp->maxcnt, NO_HINT);
  /* do some signaling here? */

  /* check if I am the last? */
  __add_and_fetch(&(bp->c2), 1);
  waitFlagMOD(&(bp->c2), bp->maxcnt, NO_HINT);
  /* do some signaling here? */

  return 0;
}

/* bar_destroy - destroy a barrier variable. */
int mytp_bar_destroy(mytp_thrMPI_Bar *bp) {
  /* nothing necessary */
  return 0 ;
}
