/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT
# include <errno.h>
# include <pthread.h>
typedef struct {
        int maxcnt;           /* maximum number of runners    */
        struct _sb {
                pthread_cond_t wait_cv;     /* cv for waiters at barrier    */
                pthread_mutex_t wait_lk;    /* mutex for waiters at barrier */
                int runners;        /* number of running threads    */
        } sb[2];
        struct _sb *sbp;      /* current sub-barrier          */
} pthread_thrMPI_Bar;

/* pthread_bar_init -  init the barrier */
int pthread_bar_init(pthread_thrMPI_Bar*bp, int count) {
  int n, i;

  if (count<1) return(EINVAL);

  bp->maxcnt = count;
  bp->sbp = &bp->sb[0];

  for (i=0; i<2; ++i) {
    struct _sb *sbp=&(bp->sb[i]);
    sbp->runners = count;

    if (n=pthread_mutex_init(&sbp->wait_lk, NULL))
      return n;

    if (n=pthread_cond_init(&sbp->wait_cv, NULL))
      return n;
  }
  return 0;
}

/* pthread_bar_wait - wait at a barrier for everyone to arrive */
int pthread_bar_wait(pthread_thrMPI_Bar *bp) {
  register struct _sb *sbp = bp->sbp;

  pthread_mutex_lock( &sbp->wait_lk );

  if ( sbp->runners == 1 ) {      /* last thread to reach barrier */
    if ( bp->maxcnt != 1 ) {
      /* reset runner count and switch sub-barriers   */
      sbp->runners=bp->maxcnt;
      bp->sbp=(bp->sbp == &bp->sb[0] )? &bp->sb[1] : &bp->sb[0];

      /* wake up the waiters */
      pthread_cond_broadcast( &sbp->wait_cv );
    }
  } 
  else {
    sbp->runners--;         /* one less runner */
    
    while ( sbp->runners != bp->maxcnt ) /* important! */
      pthread_cond_wait( &sbp->wait_cv, &sbp->wait_lk );
  }

  pthread_mutex_unlock( &sbp->wait_lk );

  return 0;
}

/* bar_destroy - destroy a barrier variable. */
int pthread_bar_destroy(pthread_thrMPI_Bar *bp) {
  int n, i;

  for (i=0; i<2; ++i ) {
    if ( n = pthread_cond_destroy( &bp->sb[i].wait_cv ) )
      return n;

    if ( n = pthread_mutex_destroy( &bp->sb[i].wait_lk ) )
      return n;
  }

  return 0 ;
}
