/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*
 * barrier.c -- BARRIER OBJECT INTERFACE 
 */

# define _REENTRANT
# include <errno.h>
# include "twait.h"

typedef struct {
        int maxcnt;           /* maximum number of runners    */
        int c1;
        int c2;
} sproc_thrMPI_Bar;

/* sproc_bar_init -  init the barrier */
int sproc_bar_init(sproc_thrMPI_Bar *bp, int count) {
  int n, i;

  if (count<1) return(EINVAL);

  bp->maxcnt = count;
  bp->c1=bp->c2=0;

  return 0;
}

/* sproc_bar_wait - wait at a barrier for everyone to arrive */
int sproc_bar_wait(sproc_thrMPI_Bar *bp
# ifdef MPISIM
	, int lseq
# endif
) {
  /* check if I am the last? */
  __add_and_fetch(&(bp->c1), 1);
  waitFlagMOD(&(bp->c1), bp->maxcnt, NO_HINT);
  /* do some signaling here? */

  /* check if I am the last? */
  __add_and_fetch(&(bp->c2), 1);
  waitFlagMOD(&(bp->c2), bp->maxcnt, NO_HINT);
  /* do some signaling here? */

  return 0;

}

/* bar_destroy - destroy a barrier variable. */
int sproc_bar_destroy(sproc_thrMPI_Bar *bp) {
  /* nothing necessary */
  return 0 ;
}
