/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*
 * barrier.c -- BARRIER OBJECT INTERFACE 
 */

# define _REENTRANT
# include <errno.h>
# include <ulocks.h>
# include "ukey.h"

#ifdef MPISIM
#include "sim.h"
#endif

extern usptr_t *thrMPI_sc_handle;

typedef struct {
# ifdef MPISIM
	barrier_t *bar2;
	int *seqno;
# endif
	barrier_t *bar;
	int maxcnt;
} sproc_thrMPI_Bar;

/* sproc_bar_init -  init the barrier */
int sproc_bar_init(sproc_thrMPI_Bar *bp, int count) {
	int n, i;

	if (count<1) return(EINVAL);

# ifdef MPISIM
	bp->bar2=new_barrier(thrMPI_sc_handle);
	init_barrier(bp->bar2);
	bp->seqno=(int *)malloc(sizeof(int)*count);
# endif

	bp->maxcnt = count;
	bp->bar=new_barrier(thrMPI_sc_handle);
	if (!bp->bar)
		return -1;

	init_barrier(bp->bar);

	return 0;
}

/* sproc_bar_wait - wait at a barrier for everyone to arrive */
int sproc_bar_wait(sproc_thrMPI_Bar *bp
# ifdef MPISIM
	, int lseq
# endif
) {
# ifdef MPISIM
	int myid=ukey_getid(), i, cur=0;
	bp->seqno[myid]=lseq;
	barrier(bp->bar2, bp->maxcnt);
	for (i=0; i<bp->maxcnt; i++) 
	  if (i!=myid) {
	    simMPI[myid].event[lseq].children[cur]=simMPI[i].event[bp->seqno[i]].seq;
            cur++;
	  }
# endif

	barrier(bp->bar, bp->maxcnt);

	return 0;
}

/* bar_destroy - destroy a barrier variable. */
int sproc_bar_destroy(sproc_thrMPI_Bar *bp) {
# ifdef MPISIM
	free_barrier(bp->bar2);
# endif
	free_barrier(bp->bar);

	return 0 ;
}
