/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*
 * barrier.c -- BARRIER OBJECT INTERFACE 
 */

# include "ukey.h"
# include "struct.h"
# include "barrier.h"
# include "comm.h"
# include "mpi.h"

#ifdef MPISIM
#include "sim.h"
#endif

/* implementation for barrier */
int MPI_Barrier(MPI_Comm communicator) {
  	thrMPI_Comm* comm;
	int myid=ukey_getid();
#ifdef MPISIM
	int lseq;
#endif

  	if ((comm=comm_verify(communicator, myid))==NULL)
		return MPI_ERR_COMM;

#ifdef MPISIM
	lseq=sim_comm_start(myid, Event_Barrier, comm, 0, 0);
#endif

  	bar_wait(comm->br
#ifdef MPISIM
		 , lseq
#endif
);

#ifdef MPISIM
	sim_comm_end(myid);
#endif
  	return MPI_SUCCESS;
}
