/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _BARRIER_H_
# define _BARRIER_H_

#define _REENTRANT

# include "thrsys.h"

# if (THREAD_SYS==PTHREAD_SYS)

# include <pthread.h>

/* barrier control object */
typedef struct {
  	int maxcnt;           /* maximum number of runners    */
  	struct _sb {
    		pthread_cond_t wait_cv;     /* cv for waiters at barrier    */
    		pthread_mutex_t wait_lk;    /* mutex for waiters at barrier */
    		int runners;        /* number of running threads    */
  	} sb[2];
  	struct _sb *sbp;      /* current sub-barrier          */
} pthread_thrMPI_Bar;

# define thrMPI_Bar pthread_thrMPI_Bar 
# define bar_init pthread_bar_init
# define bar_wait pthread_bar_wait
# define bar_destroy pthread_bar_destroy

# elif (THREAD_SYS==SPROC_SYS)

/**
# include <ulocks.h>
typedef struct {
# ifdef MPISIM
	barrier_t *bar2;
	int *seqno;
# endif
	barrier_t *bar;
	int maxcnt;
} sproc_thrMPI_Bar;
**/

typedef struct {
        int maxcnt;           /* maximum number of runners    */
	int c1;
	int c2;
} sproc_thrMPI_Bar;

# define thrMPI_Bar sproc_thrMPI_Bar
# define bar_init sproc_bar_init
# define bar_wait sproc_bar_wait
# define bar_destroy sproc_bar_destroy

# else

/* barrier control object */
typedef struct {
        int maxcnt;           /* maximum number of runners    */
	int c1;
	int c2;
} mytp_thrMPI_Bar;

# define thrMPI_Bar mytp_thrMPI_Bar
# define bar_init mytp_bar_init
# define bar_wait mytp_bar_wait
# define bar_destroy mytp_bar_destroy

# endif

int bar_init(thrMPI_Bar *bp, int count);
int bar_wait(thrMPI_Bar *bp
#ifdef MPISIM
,int
#endif
);
int bar_destroy(thrMPI_Bar *bp);

# endif
