/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _MQUEUE_H_
# define _MQUEUE_H_

#define _REENTRANT

# include "tqueue.h"
# include "struct.h"
# include "mpi.h"

typedef struct _thrMPI_Channel {
	TQueue *rq;
	TQueue *sq;

# if (_SR_CHECK_)
	int s_no; /* the send seq no for messages go to this channel */
	int r_no; /* the receive seq no for messages go to this channel */
# endif

# if (_LOCK_)
	/* lock for channel access */
	int lock_flag;
# endif 
} thrMPI_Channel;

# define INUSE_FREE_UP 10 /* where there are 10% free handles, GC is triggered */
# define INUSE_FREE_LOW 20 /* where there are 5% free handles, GC is stopped */

/* in any case, GC will not continue when there are no more than 5 free handles */
# ifndef GC_MIN
# define GC_MIN 5 
# endif

extern thrMPI_Channel **thrMPI_GCH; /* global channel array */

#if (!_LOCK_)
/* by sender */
thrMPI_RHandle *rq_probe(TQueue *q, int context, int tag, int *ts);
/* by receiver */
thrMPI_SHandle *sq_probe(TQueue *q, int context, int tag, int *ts);
#else
/* by sender */
thrMPI_RHandle *rq_probe(TQueue *q, int context, int tag);
/* by receiver */
thrMPI_SHandle *sq_probe(TQueue *q, int context, int tag);
#endif

thrMPI_RHandle *rh_new(TQueue *q);
thrMPI_SHandle *sh_new(TQueue *q);

void rh_delete(TQueue *q, thrMPI_RHandle *rh);
void sh_delete(TQueue *q, thrMPI_SHandle *sh);

/* Garbage collection */
# if (_GC_)
/* by receiver */
void rq_GC(TQueue *rq);
/* by sender */
void sq_GC(TQueue *sq);
# endif

void user_GC(int myid);

/* I guess set chunk_size to be equal to group_size is a proper choice */
TQueue *thrMPI_sq_create(int chunk_size, int owner);
TQueue *thrMPI_rq_create(int chunk_size, int owner);

int thrMPI_GCH_ginit(void); /* init the whole channel structure */
int thrMPI_GCH_uinit(int myid); /* init the queues owned by the process */

int thrMPI_SH_init(MPI_Comm comm, char *buf, int len, int from, int to, int tag, int count, thrMPI_Mode mode, Boolean isPersistent, Boolean isImmediate, thrMPI_Request *psh);

int thrMPI_RH_init(MPI_Comm comm, char *buf, int len, int from, int to, int tag, MPI_Datatype datatype, int count, Boolean isPersistent, Boolean isImmediate, thrMPI_Request *prh);

/* these are the core functions for send/recv family functions */
/* not for standard send/recv */
int thrMPI_Start(thrMPI_Request *request); 

/* int thrMPI_Test(thrMPI_Request *request, int *flag, thrMPI_Status *status); */

int thrMPI_Waitall( int, MPI_Request*, MPI_Status*);
int thrMPI_Wait(thrMPI_Request *request, thrMPI_Status *status);

int thrMPI_Request_free(thrMPI_Request *request);

/* not for standard send/recv */
int thrMPI_Cancel(thrMPI_Request *reqest);

/* int thrMPI_Test_cancelled(thrMPI_Status *status, int *flag); */
/* thrMPI_Waitall thrMPI_Waitany, thrMPI_Waitsome, thrMPI_Freeall thrMPI_Freeany, thrMPI_Freesome are easy */
/* to implement (no intention to complete them at this stage) */

# endif
