/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _REDUCE_H_
# define _REDUCE_H_

#define _REENTRANT
# include "ukey.h"
# include "mpi.h"

typedef struct _thrMPI_RDCHandle {
	struct _thrMPI_RDCHandle *next;
	int avail; /* -1: last handle in the queue, 0: not available, 1: available */
	void *data;
	unsigned int count; /* in the unit of datatype */
} thrMPI_RDCHandle;
	
typedef struct _thrMPI_RDCQueue {
	ukey_t key_curhandle; /* each process keeps the current handle pointer */
	struct _thrMPI_RDCHandle *free;
	struct _thrMPI_RDCHandle *first;
	struct _thrMPI_RDCHandle cache[1]; /* a variable length array */
} thrMPI_RDCQueue;	

typedef struct _thrMPI_ReduceObj {
	int group_size;
	thrMPI_RDCQueue *(rdco_array[1]); /* indexed by the root, variable length */
} thrMPI_ReduceObj;
	
/* reduce all object */
typedef struct _thrMPI_ARDCObj {
  int group_size;
  struct _sub {
    int avail;
    int done;
    void *data;
    unsigned int count;
  } sub[2];
  struct _sub *subp;
} thrMPI_AllreduceObj;

thrMPI_ReduceObj *thrMPI_rdco_create(int groupsize); /* constructor */
int thrMPI_rdco_destroy(thrMPI_ReduceObj *rdco); /* destructor */
int thrMPI_rdco_reduce(thrMPI_ReduceObj *rdco, int myid, int root, void *sendbuf, void *recvbuf, unsigned int count, MPI_Datatype datatype, MPI_Op operator);

thrMPI_AllreduceObj *thrMPI_ardco_create(int groupsize); /* constructor */
int thrMPI_ardco_destroy(thrMPI_AllreduceObj *ardco); /* destructor */
int thrMPI_ardco_allreduce(thrMPI_AllreduceObj *ardco, int myid, void *sendbuf, void *recvbuf, unsigned int count, MPI_Datatype datatype, MPI_Op operator);

# endif
