/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#ifdef MPISIM

#ifndef _sim_h_
#define _sim_h_

#include "mpi.h"

/* sim.h - data structures and etc for MPI simulation */

#define _REENTRANT

enum {
  Event_Compute=0, 
  Event_Send=1, 
  Event_Recv=2, 
  Event_Bcast=3, 
  Event_Barrier=4
};

typedef struct {
  int seq;		/* sequence number of the event,
			   higher 16 bits are host number and lower 16 bits start
			   from 0 on each host */
  int type;		/* type of the event */
  int size;		/* size of the data communicated */
  double exec_time;	/* the time for the execution of the event itself */
  double start_time;	/* starting time of the event, decided off-line */
  double block_time;	/* the time for blocking, if it is a block operation,
			   decided off-line */
  int parents;		/* number of events it needs to wait, 0 means non-blocking op,
			   changed off-line */
  int child_num;	/* number of events it needs to wake up */
  int* children;
} simMPI_Event;

typedef struct {
  int MAX_EVENT_NUM;
  int EventNum;
  simMPI_Event *event;
  double clock;
} simMPI_ctrl_t;

extern simMPI_ctrl_t* simMPI;

/* function declarations */
void sim_global_init(int);
void sim_init(int);
void sim_gstart(int);
int sim_comm_start(int, int, MPI_Comm, int, int);
void sim_comm_end(int);
void sim_gend(int);
void sim_analyze(int);

/* architecture-dependent communication parameters */
#define ALPHA 0.00002		/* startup time */
#define BELTA 1.23e-8
#define COMPUTE 1		/* computation scales */

#endif

#endif
