/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _SYNC_H_
# define _SYNC_H_

#define _REENTRANT

# include "thrsys.h"

/* Interface: mutex lock: mtx_create, mtx_init, mtx_trylock, mtx_lock, mtx_unlock, mtx_destroy.
 * Note: mtx_create includes mtx_init.
 */

# if (THREAD_SYS==PTHREAD_SYS)
# include <pthread.h>
typedef pthread_mutex_t *pthread_mtx_t;
# define mtx_t pthread_mtx_t
# define mtx_create pthread_mtx_create
# define mtx_init pthread_mtx_init
# define mtx_trylock pthread_mutex_trylock
# define mtx_lock pthread_mutex_lock
# define mtx_unlock pthread_mutex_unlock
# define mtx_destroy pthread_mtx_destroy

# else

# include <abi_mutex.h>
typedef abilock_t *sproc_mtx_t;
# define mtx_t sproc_mtx_t

# define mtx_create sproc_mtx_create
# define mtx_init init_lock
# define mtx_trylock acquire_lock
# define mtx_lock sproc_mtx_lock
# define mtx_unlock release_lock
# define mtx_destroy sproc_mtx_destroy

# endif

extern mtx_t mtx_create(int myid);
extern int mtx_init(mtx_t mtx);
extern int mtx_trylock(mtx_t mtx);
extern int mtx_lock(mtx_t mtx);
extern int mtx_unlock(mtx_t mtx);
extern int mtx_destroy(mtx_t mtx, int myid);

# endif
