/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _TWAIT_H_
# define _TWAIT_H_

#define _REENTRANT
# define _BSD_SIGNALS
# include <signal.h>
# include "thrsys.h"

# if (THREAD_SYS!=MYTP_SYS)
# define VOLATILE volatile
# else
# define VOLATILE
# endif

/**
# define VOLATILE volatile
**/

# define NO_HINT -1

enum {thrMPI_IDLE=0, thrMPI_BUSY=1, thrMPI_SLEEP=2}; 

typedef struct _thrMPITCB {
	pid_t pid;
	int intr; /* interrupt flag */
	int status; /* thread status, may not be accurate */
} thrMPITCB;

void twait_ginit();
void twait_uinit(int myid);

/* not really important functions, just memory stuff */

void twait_uend(int myid);
void twait_gend();

/* a set of waitFlag functions and signalFlag functions */

/* extern thrMPITCB **thrMPI_tcbs; */

# define ROUNDUP(val, intvl) (((val)-1)/(intvl)+1)*(intvl)

/* the following parameters are architecture dependent, can be obtained by 
 * running the kernel benchmarks in $(MPI_SRC_HOME)/kernelbench */
# define BUSY_START_IT (32) /* 64 busy iterations takes about 1us, but considering overhead, take half. */
# define BUSY_START (1)  /* in the unit of 1us, same below */
# define BUSY_END_IT (65536*4) /* 65536 busy iterations takes about 1ms */
# define BUSY_END (1024*4) 

/* nanosleep has a resolution of 0.8ms but starts at least 1.6ms */
# define NSLP_START_IT (1600000)
# define NSLP_START (1600)
# define NSLP_END_IT (1000000000)
# define NSLP_END  (1000000) /* 1 second */
# define NSLP_INT_IT (800000) /* nanosleep interval */
# define NSLP_INT (800)

/* usleep has a resoltuion of 10ms and starts at 10ms */
/* usleep will not be used because it will cause process dead! */

/* sleep will be used when >1sec */

/* from BUSY_START to EXP_END, a exponential back-off will be used */
/* EXP_END must always be no less than BUSY_END */
# define EXP_END (10000) 
/* from EXP_END to LIN_END, a linear back-off will be used */
# define LIN_END (100000) 
# define LIN_INT (1000) /* incr 1ms each iteration */

/* beyond LIN_END, a fixed interval probe will be used */
# define FIX_INT (100000) /* fixed interval */

/* yield cost, need to be tuned */
# define YIELD_COST (1)
# define YIELD_COST_IT ((1*64))
/* When busy waiting time is at least two times the YIELD cost, add one sched_yield between 
 * busy waiting.
 */

/* waitFlagOP will block until the relationship (*pflag OP value) is true 
 * It uses an adaptive algorithm that automatically shift from busy testing to 
 * sleep (nanosleep is used for the best resolution).
 * Assuming there is only one writer to the flag and could be multiple readers.
 */
int waitFlagEQ(VOLATILE int *pflag, int value, int hint);
int waitFlagNE(VOLATILE int *pflag, int value, int hint);
int waitFlagLE(VOLATILE int *pflag, int value, int hint);
int waitFlagIN(VOLATILE int *pflag, int value, int hint);
int waitFlagMOD(VOLATILE int *pflag, int value, int hint);

/* signalFlag will set the flag to value and wakeup the sleeping process */
/* Note the process to be signaled can be either busy waiting or sleeping */
int signalFlag(int id, int *pflag, int value);
int fSignalFlag(int id, unsigned int *pflag, int value);
int signalConditionalFlag(int id, int *pflag, int value, int oldval);
int signalAllFlag(int *pflag, int value);
int signalAllConditionalFlag(int *pflag, int value, int oldval);

# endif
