/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <stdio.h>
# include <sys/types.h>
# include <sys/prctl.h>
# include <abi_mutex.h>
# include <sys/sysmp.h>
# include <sys/sysinfo.h> 

# define NT 1000000
# define PNO 16

double gettime();
int turn=0;

abilock_t lck, lck2;

void test(void *arg)
{
	double t1, t2;
	int i=0;

	sysmp(MP_MUSTRUN_PID, PNO+1, getpid());

	spin_lock(&lck);
	t1=gettime();
	for (;i<NT/100-1;i++) {
		release_lock(&lck2);
		spin_lock(&lck);
	}	
	release_lock(&lck2);
	t2=gettime();
		
	printf("//release and spin lock =%e\n", (t2-t1)/(NT/100-1));

	return ;
}

main()
{
	double t1, t2;
	int i;

	sysmp(MP_MUSTRUN_PID, PNO, getpid());
	init_lock(&lck);
	init_lock(&lck2);
	acquire_lock(&lck);
	
	t1=gettime();
	for (i=0;i<NT;i++)
		acquire_lock(&lck);
	t2=gettime();
	printf("acquire_lock=%e\n", (t2-t1)/NT);
	
	t1=gettime();
	for (i=0;i<NT;i++)
		release_lock(&lck);
	t2=gettime();
	printf("release_lock=%e\n", (t2-t1)/NT);

	t1=gettime();
	for (i=0;i<NT;i++) {
		acquire_lock(&lck);
		release_lock(&lck);
	}
	t2=gettime();
	printf("acquire then release lock=%e\n", (t2-t1)/NT);

	acquire_lock(&lck); /* make sure the lock has been held by the program */
	acquire_lock(&lck2); /* make sure the lock has been held by the program */

	sproc(test, PR_SADDR, NULL);
	sleep(1); /* make sure the child already starts running */

	t1=gettime();
	for (i=0;i<NT/100;i++) {
		release_lock(&lck);
		spin_lock(&lck2);
	}
	t2=gettime();
	printf("release and spin lock =%e\n", (t2-t1)/NT);
	
	return 0;
}
