/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*************************************************/

# ifndef _DEBUG_
# define _DEBUG_ 1
# endif

# define _REENTRANT

/*****************
 * MYTP version *
 *****************/
# include <strings.h>
# include <stdio.h>
# include <stdlib.h>
# include "mytp.h"

/**
 * The following two variables are shared and could be modified by all
 * threads in the shared group 
 */
typedef int mytp_ukey_t;

static mytp_ukey_t nkey=0; /* next available key */

# define KEY_MAX 1024
# define ID_MAX 1000

static int mytp_key_max=KEY_MAX;

int mytp_ukey_thrcreate(void (*start)(void), short id)
{
	int prs;

	if ((id<0)||(id>mytp_key_max)) {
# if _DEBUG_
		fprintf(stderr, "Illegal ID\n");
# endif
		return 0;
	}

	if ((prs=mytp_create( (mytp_userf_t *)start, NULL))<0) {
		fprintf(stderr, "Failed to create MyTP!\n");
		return NULL;
	}
	else {
# if _DEBUG_
		if (id!=prs)
			fprintf(stderr, "Incompatible thread creation convention!\n");
# endif
		return prs;
	}
}

int mytp_ukey_createkey(mytp_ukey_t *pkey)
{
	register mytp_ukey_t tmp;

	tmp=__add_and_fetch(&nkey, 1); 

	if ( tmp>mytp_key_max )
		return -1;

	*pkey=tmp;
	return 0;
}

int mytp_ukey_setval(mytp_ukey_t key, void *val)
{
	register void **ukey_vals=mytp_get_private();

	/**
	fprintf(stderr, "%d Got private %x\n", mytp_get_myid(), ukey_vals);
	fflush(stderr);
	**/

	if ( (key<=0) || (key>mytp_key_max) )
		return -1;
	
	ukey_vals[key-1]=val;

	return 0;
}

void * mytp_ukey_getval(mytp_ukey_t key)
{
	register void **ukey_vals=mytp_get_private();

	/**
	fprintf(stderr, "%d Got private %x\n", mytp_get_myid(), ukey_vals);
	fflush(stderr);
	**/

	/* It is kind of dangerous to return a valid value on error. */
	if ( (key<=0) || (key>mytp_key_max) )
		return NULL;

	return ukey_vals[key-1];
}
