/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*************************************************/

# ifndef _DEBUG_
# define _DEBUG_ 1
# endif

# define _REENTRANT

/*******************
 * PTHREAD version *
 *******************/

# include <stdio.h>
# include <stdlib.h>
# include <pthread.h>

# define ID_MAX 1000

typedef struct parm {
	void (*start)(void);
	short id;
} PARM;

void pthread_ukey_setid(short id);

static void *pthread_ukey_stub(void *parm)
{
	pthread_ukey_setid( ((PARM *)parm)->id );
	(* ((PARM *)parm)->start)();
	free (parm);
	return NULL;
}
	
static pthread_key_t pthread_idkey=0; /* 0 means not yet initialized */
pthread_t *pthread_ukey_thrcreate(void (*start)(void), short id)
{
	PARM *parm;
	pthread_t *thread;
	int rt;

	if ((id<0)||(id>ID_MAX))
		return NULL;
	
	parm=(PARM *)malloc(sizeof(PARM));
	parm->start=start;
	parm->id=id;
	
	thread=(pthread_t *)malloc(sizeof(pthread_t));
	if (rt=pthread_create(thread, NULL, pthread_ukey_stub, (void *)parm))
		return NULL;
	else 
		return thread;
}

void pthread_ukey_setid(short id)
{
	pthread_key_t newkey;

	if (pthread_idkey==0) { /* hasn't been set */
		pthread_key_create(&newkey, NULL);
		if (!__compare_and_swap(&pthread_idkey, (pthread_key_t)0, newkey)) 
		/* idkey has already been set */
			pthread_key_delete(newkey);
	}

	pthread_setspecific(pthread_idkey, (void *)(long)(unsigned short)id);
	return;
}

short pthread_ukey_getid(void)
{
	return (short)(long)pthread_getspecific(pthread_idkey);
}
