/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*************************************************/

# ifndef _DEBUG_
# define _DEBUG_ 1
# endif

# define _REENTRANT

/*****************
 * SPROC version *
 *****************/
# include <strings.h>
# include <stdio.h>
# include <stdlib.h>
/**
# include <abi_mutex.h>
**/

/**
 * The following two variables are shared and could be modified by all
 * threads in the shared group 
 */
typedef int sproc_ukey_t;

static sproc_ukey_t nkey=0; /* next available key */
/**** The following lock is not used now. Atomic addition is used as a
 more efficient implementation. */
/* static abilock_t nkey_lck; */ /* since static are all initialized to 0 which is 
 	the unlocked state of abilock, init_lock is not necessary */

# define KEY_MAX 1024
# define ID_MAX 1000

static int sproc_key_max=KEY_MAX;

# include <sys/types.h>
# include <sys/resource.h>
# include <sys/prctl.h>

# define UKEY_PRDA_MAGIC_MSK (-((signed long)((unsigned short)-1)+1))
# define UKEY_PRDA_ID_MSK ((signed  long)((unsigned short)-1))
# define UKEY_PRDA_MAGIC ( (((unsigned long)-1)/3*2)&UKEY_PRDA_MAGIC_MSK)
# define USER_PRDA_SIZE sizeof(PRDA->usr_prda.fill)

# define NUM_KEY1 ((USER_PRDA_SIZE)/sizeof(void *)-2)
# define NUM_KEY2 (sproc_key_max)

typedef struct ukey_prda_user {
	signed long ukey_magicid; /* same space as void * */
	void * ukey_key1[NUM_KEY1];
	void ** ukey_key2;
} ukey_prda_user_t;

# define UKEY_PRDA_USER ((ukey_prda_user_t *)(PRDA->usr_prda.fill))
# define UKEY_GETID ( (short)((UKEY_PRDA_USER->ukey_magicid)&UKEY_PRDA_ID_MSK) )
# define UKEY_GETMAGIC (UKEY_PRDA_USER->ukey_magicid&UKEY_PRDA_MAGIC_MSK)
# define UKEY_MAKEID(id) (UKEY_PRDA_MAGIC|(signed long)(unsigned short)(id))

typedef struct parm {
	void (*start)(void);
	short id;
} PARM;

void sproc_ukey_setid(short id);

static void sproc_ukey_stub(void *parm)
{
	sproc_ukey_setid( ((PARM *)parm)->id );
	(* ((PARM *)parm)->start)();
	free (parm);
	return;
}
	
pid_t sproc_ukey_thrcreate(void (*start)(void), short id)
{
	PARM *parm;
	pid_t prs;

	if ((id<0)||(id>ID_MAX)) {
# if _DEBUG_
		fprintf(stderr, "Illegal ID\n");
# endif
		return 0;
	}

	parm=(PARM *)malloc(sizeof(PARM));
	parm->start=start;
	parm->id=id;
	
	/** printf("PR_SADDR=%d\n", PR_SADDR); **/
	if ((prs=sproc(sproc_ukey_stub, PR_SADDR, (void *)parm))<0) {
		perror("");
		return NULL;
	}
	else
		return prs;
}

void sproc_ukey_setid(short id)
{
# if _DEBUG_
	/* it is possible for a proc that spawns thread after calling setid */
	if ((UKEY_GETMAGIC!=0) || (UKEY_GETID!=0)) {
		fprintf(stderr,"DEBUG, in setid, ID has already been set or data structure corrupted.\n");
		/* return; */
	}
# endif
	if (id>=ID_MAX) {
		fprintf(stderr,"ID out of range (0-999).\n");
		return;
	}
		
	UKEY_PRDA_USER->ukey_magicid=UKEY_MAKEID(id);

	return;
}

short sproc_ukey_getid(void)
{
	short id;

	if ( (UKEY_GETMAGIC!=UKEY_PRDA_MAGIC) || ( (id=UKEY_GETID)>=ID_MAX) ) {
# if _DEBUG_
		fprintf(stderr,"DEBUG, in getid, user PRDA data structure corrupted.\n");
# endif
		return -1;
	}

	return id;
}

int sproc_ukey_createkey(sproc_ukey_t *pkey)
{
	register sproc_ukey_t tmp;

	/** spin_lock(&nkey_lck); **/
	tmp=__add_and_fetch(&nkey, 1); 
	/** release_lock(&nkey_lck); **/

	if ( tmp>(NUM_KEY1+NUM_KEY2) )
		return -1;

	*pkey=tmp;
	return 0;
}

int sproc_ukey_setval(sproc_ukey_t key, void *val)
{
	if ( (key<=0) || (key>(NUM_KEY1+NUM_KEY2)) )
		return -1;

# if _DEBUG_
	if ( (UKEY_GETMAGIC!=UKEY_PRDA_MAGIC) || (UKEY_GETID>=ID_MAX) ) {
		fprintf(stderr,"DEBUG, in setval, user PRDA data structure corrupted.\n");
		return -1;
	}
# endif

	if (key>NUM_KEY1) {
		if (UKEY_PRDA_USER->ukey_key2==NULL) { /* need to expand */
			UKEY_PRDA_USER->ukey_key2=(void **)malloc(sizeof(void *)*NUM_KEY2);
			bzero(UKEY_PRDA_USER->ukey_key2, sizeof(void *)*NUM_KEY2);
		}
		UKEY_PRDA_USER->ukey_key2[key-NUM_KEY1]=val;
	}
	else
		UKEY_PRDA_USER->ukey_key1[key-1]=val;
	
	return 0;
}

void * sproc_ukey_getval(sproc_ukey_t key)
{
	/* It is kind of dangerous to return a valid value on error. */
	if ( (key<=0) || (key>(NUM_KEY1+NUM_KEY2)) )
		return NULL;

# if _DEBUG_
	if ( (UKEY_GETMAGIC!=UKEY_PRDA_MAGIC) || (UKEY_GETID>=ID_MAX) ) {
		fprintf(stderr,"DEBUG, in getval, user PRDA data structure corrupted.\n");
		return NULL;
	}
# endif

	if (key>NUM_KEY1) {
		if (UKEY_PRDA_USER->ukey_key2==NULL) { /* need to expand */
			UKEY_PRDA_USER->ukey_key2=(void **)malloc(sizeof(void *)*NUM_KEY2);
			bzero(UKEY_PRDA_USER->ukey_key2, sizeof(void *)*NUM_KEY2);
		}
		return UKEY_PRDA_USER->ukey_key2[key-NUM_KEY1];
	}
	else
		return UKEY_PRDA_USER->ukey_key1[key-1];
}

/**
 * Calling ukey_setkeymax after more than NUM_KEY1 keys are created will 
 * be of no use in this version. 
 */
void ukey_setkeymax(int max)
{
	if (UKEY_PRDA_USER->ukey_key2) /* portion 2 has been allocated */
		return;

	sproc_key_max=max;
}
