/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT
# include <pthread.h>
# include "anl_shmem.h"

/* prototypes */
pthread_mutex_t *pthread_mtx_create(int myid);
int pthread_mtx_init(pthread_mutex_t *mtx);
int pthread_mtx_destroy(pthread_mutex_t *mtx, int myid);

pthread_mutex_t * pthread_mtx_create(int myid)
{
	pthread_mutex_t *mtx;

	if ((mtx=anl_malloc(myid, sizeof(pthread_mutex_t)))==NULL)
		return NULL;

	if (pthread_mtx_init(mtx)) {
		anl_free(myid, mtx);
		return NULL;
	}

	return mtx;
}

int pthread_mtx_init(pthread_mutex_t *mtx)
{
	pthread_mutexattr_t attr;
	
	/* priority inversion */
	pthread_mutexattr_setprotocol(&attr, PTHREAD_PRIO_INHERIT);
	pthread_mutexattr_setprioceiling(&attr, 0); 

	if (pthread_mutex_init(mtx, &attr))
		return -1;
	
	return 0;
}

int pthread_mtx_destroy(pthread_mutex_t *mtx, int myid)
{
	if (pthread_mutex_destroy(mtx))
		return -1;

	anl_free(myid, mtx);

	return 0;
}
