/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <stdio.h>
# include "anl_shmem.h"
# include "tqueue.h"
# include "ukey.h"

# define NT 1000000
TQueue *p;
void printQ(TQueue *queue);
void Q_status(TQueue *queue);
double gettime();
struct s {
	int i;
} *e;

# define ASSERT(s) { if (!(s)) fprintf(stderr, "assertion failed\n"); }

extern int maxQlen;

int comp(void *f, int context, int tag, int arg, void *arg2)
{
	return ( ((struct s *)f)->i-tag);
}
int thrMPI_Nproc=1;
main()
{
	int c;
	int i=0;
	int last=1;
	int eq=0, dq=0, pq=0, spq=0, fq=0, sfq=0;
	double t1, t2;

	anl_shmem_ginit();
	ukey_setid(0);
	anl_shmem_uinit(0, 1000000);
	srand((int)gettime());

	p=Q_create(comp, sizeof(struct s), 8);

	
	t1=gettime();
	for (;i<NT;i++) {
		c=rand()%8;
		switch (c) {
		case 0:
		case 1:
		case 2: /* enQ */
			e=(struct s *)Q_newElem(p);
			e->i=last;
			last=(last+1)>>1;
			enQ(p, e);
			eq++;
			break;
		case 3:
		case 4:
		case 5:
			e=deQ(p);
			if (e)
				Q_freeElem(p, e);
			dq++;
			break;
		case 6:
			e=Q_probe(p, 0, 2, 0, (void *)0);
			if (!e) {
				pq++;
				last=2;
			}
			else
				spq++;
			break;
		case 7:
			e=Q_find(p, 0, 2, 0, (void *)0);
			if (!e) {
				fq++;
				last=2;
			}
			else {
				sfq++;
				Q_freeElem(p, e);
			}
			break;
		}
	}
	t2=gettime();
	printf("average cost=%e\n", (t2-t1)/NT);
	printf("enQ=%d\ndeQ=%d\nfailed Q_prob=%d\nsuccessful Q_probe=%d\nfailed Q_find=%d\nsuccessful Q_find=%d\n", eq, dq, pq, spq, fq, sfq);
	printf("max queue length=%d\n", maxQlen);
	Q_status(p);
	anl_shmem_uend(0);
	anl_shmem_gend();
}
