/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT

# define _BSD_SIGNALS
# include <time.h>
# include "twait.h"
# include "ukey.h"

# define BUSY_RES 1 /* in the unit of YIELD_COST */

int thrMPI_donotopt=0; /* a trash variable */
extern int thrMPI_Nproc;

# if (_DEBUG_)
int busy=0, sigs=0, prob=0;
# endif

static void foo(void)
{ 

	/* printf("intr from sig handle\n"); */
# if (_DEBUG_)
	sigs++;
# endif
	return; 
}

# define OP ==
int waitFlagEQ(VOLATILE int *pflag, int value, int hint)
{
	int myid;
	int *pintr;
	register int long_term=0;

	if (*pflag OP value) return 0;

	if ((myid=ukey_getid())==-1) return -1;
	pintr=&(thrMPI_tcbs[myid]->intr);

	thrMPI_tcbs[myid]->intr=0; /* clear interrupt */
	thrMPI_tcbs[myid]->status=thrMPI_BUSY; /* busy waiting */

	if (*pflag OP value) {
# if (_DEBUG_)
		busy++;
# endif
		thrMPI_tcbs[myid]->status=thrMPI_IDLE;
		return 0;
	}

	while (1) {
		volatile int i=YIELD_COST*BUSY_RES;
		
		long_term=(long_term+1)%1000;
		if (!long_term) {
			sginap(1);
		}
		
		if (!(*pflag OP value))
			while ((i--) && (!(*pintr)) )
				thrMPI_donotopt++; 

		if (*pflag OP value) {
			/* printf("intr from busy waiting\n"); */
# if (_DEBUG_)
			busy++;
# endif
			thrMPI_tcbs[myid]->status=thrMPI_IDLE;
			return 0;
		}
		else if (!i) /* if i==0, then clear the interrupt */
			*pintr=0; /* false interrupt */
	}

	return 0;
}
# undef OP

# define OP !=
int waitFlagNE(VOLATILE int *pflag, int value, int hint)
{
	int myid;
	int *pintr;
	register int long_term=0;

	if (*pflag OP value) return 0;

	if ((myid=ukey_getid())==-1) return -1;
	pintr=&(thrMPI_tcbs[myid]->intr);

	thrMPI_tcbs[myid]->intr=0; /* clear interrupt */
	thrMPI_tcbs[myid]->status=thrMPI_BUSY; /* busy waiting */

	if (*pflag OP value) {
# if (_DEBUG_)
		busy++;
# endif
		thrMPI_tcbs[myid]->status=thrMPI_IDLE;
		return 0;
	}

	while (1) {
		volatile int i=YIELD_COST*BUSY_RES;
		
		long_term=(long_term+1)%1000;
		if (!long_term) {
			sginap(1);
		}
		
		if (!(*pflag OP value))
			while ((i--) && (!(*pintr)) )
				thrMPI_donotopt++; 

		if (*pflag OP value) {
			/* printf("intr from busy waiting\n"); */
# if (_DEBUG_)
			busy++;
# endif
			thrMPI_tcbs[myid]->status=thrMPI_IDLE;
			return 0;
		}
		else if (!i) /* if i==0, then clear the interrupt */
			*pintr=0; /* false interrupt */
	}

	return 0;
}

# undef OP

# define OP <=
int waitFlagLE(VOLATILE int *pflag, int value, int hint)
{
	int myid;
	int *pintr;
	register int long_term=0;

	if (*pflag OP value) return 0;

	if ((myid=ukey_getid())==-1) return -1;
	pintr=&(thrMPI_tcbs[myid]->intr);

	thrMPI_tcbs[myid]->intr=0; /* clear interrupt */
	thrMPI_tcbs[myid]->status=thrMPI_BUSY; /* busy waiting */

	if (*pflag OP value) {
# if (_DEBUG_)
		busy++;
# endif
		thrMPI_tcbs[myid]->status=thrMPI_IDLE;
		return 0;
	}

	while (1) {
		volatile int i=YIELD_COST*BUSY_RES;
		
		long_term=(long_term+1)%1000;
		if (!long_term) {
			sginap(1);
		}
		
		if (!(*pflag OP value))
			while ((i--) && (!(*pintr)) )
				thrMPI_donotopt++; 

		if (*pflag OP value) {
			/* printf("intr from busy waiting\n"); */
# if (_DEBUG_)
			busy++;
# endif
			thrMPI_tcbs[myid]->status=thrMPI_IDLE;
			return 0;
		}
		else if (!i) /* if i==0, then clear the interrupt */
			*pintr=0; /* false interrupt */
	}

	return 0;
}

# undef OP

# define OP &
/* wait until (*pflag&value)!=0 */
int waitFlagIN(VOLATILE int *pflag, int value, int hint)
{
	int myid;
	int *pintr;
	register int long_term=0;

	if (*pflag OP value) return 0;

	if ((myid=ukey_getid())==-1) return -1;
	pintr=&(thrMPI_tcbs[myid]->intr);

	thrMPI_tcbs[myid]->intr=0; /* clear interrupt */
	thrMPI_tcbs[myid]->status=thrMPI_BUSY; /* busy waiting */

	if (*pflag OP value) {
# if (_DEBUG_)
		busy++;
# endif
		thrMPI_tcbs[myid]->status=thrMPI_IDLE;
		return 0;
	}

	while (1) {
		volatile int i=YIELD_COST*BUSY_RES;
		
		long_term=(long_term+1)%1000;
		if (!long_term) {
			sginap(1);
		}
		
		if (!(*pflag OP value))
			while ((i--) && (!(*pintr)) )
				thrMPI_donotopt++; 

		if (*pflag OP value) {
			/* printf("intr from busy waiting\n"); */
# if (_DEBUG_)
			busy++;
# endif
			thrMPI_tcbs[myid]->status=thrMPI_IDLE;
			return 0;
		}
		else if (!i) /* if i==0, then clear the interrupt */
			*pintr=0; /* false interrupt */
	}

	return 0;
}

# undef OP
	
/* Unsafe signal, if issuing two signals to the same flag, 
 * it is possible to lose the previous one.
 */ 
int signalFlag(int id, int *pflag, int value)
{
	register int status;

# if (_DEBUG_)
	if (id<0 || id>thrMPI_Nproc) return -1; 
# endif
	*pflag=value;

	status=thrMPI_tcbs[id]->status; 
	
	if (status==thrMPI_BUSY) {
		thrMPI_tcbs[id]->intr=1; /* set interrupt */
	}
	else if (status==thrMPI_SLEEP) {
		kill(thrMPI_tcbs[id]->pid, SIGUSR1); 
	}
	else { /* must be a problem */
# if (_DEBUG_)
		prob++;
# endif
	}

	return 0;
}

/* signal the flag only when the flag is equal to oldval */
int signalConditionalFlag(int id, int *pflag, int value, int oldval)
{
	register int status;

# if (_DEBUG_)
	if (id<0 || id>thrMPI_Nproc) return -1;
# endif

	if (*pflag!=oldval) return -1;
	if (!__compare_and_swap(pflag, oldval, value))
		return -1;

	status=thrMPI_tcbs[id]->status; 
	
	if (status==thrMPI_BUSY) {
		thrMPI_tcbs[id]->intr=1; /* set interrupt */
	}
	else if (status==thrMPI_SLEEP) {
		kill(thrMPI_tcbs[id]->pid, SIGUSR1); 
	}
	else { /* must be a problem */
# if (_DEBUG_)
		prob++;
# endif
	}

	return 0;
}

# if 0 
int signalAllFlag(int *pflag, int value)
{
	register int status;
	int myid;
	int id=0;

	*pflag=value;

	myid=ukey_getid();

	for (; id<thrMPI_Nproc; id++) {
		if (id==myid) continue;

		status=thrMPI_tcbs[id]->status; 
	
		if (status==thrMPI_BUSY) {
			thrMPI_tcbs[id]->intr=1; /* set interrupt */
		}
		else if (status==thrMPI_SLEEP) {
			kill(thrMPI_tcbs[id]->pid, SIGUSR1); 
		}
		else { /* must be a problem */
# if (_DEBUG_)
			prob++;
# endif
		}
	}

	return 0;
}

/* signal the flag only when the flag is equal to oldval */
int signalAllConditionalFlag(int *pflag, int value, int oldval)
{
	register int status;
	int id=0, myid;

	myid=ukey_getid();

	if (*pflag!=oldval) return -1;

	if (!__compare_and_swap(pflag, oldval, value))
		return -1;

	for (; id<thrMPI_Nproc; id++) {
		if (id==myid) continue;

		status=thrMPI_tcbs[id]->status; 
		
		if (status==thrMPI_BUSY) {
			thrMPI_tcbs[id]->intr=1;
		}
		else if (status==thrMPI_SLEEP) {
			kill(thrMPI_tcbs[id]->pid, SIGUSR1); 
		}
		else { /* must be a problem */
# if (_DEBUG_)
			prob++;
# endif
		}
	}

	return 0;
}

# endif

int signalAllFlag(int *pflag, int value)
{
	register int status;
	int myid;
	int id=0;

	*pflag=value;

	myid=ukey_getid();

	for (; id<thrMPI_Nproc; id++) {
		if (id==myid) continue;

		status=thrMPI_tcbs[id]->status; 
	
		if (status==thrMPI_BUSY) {
			thrMPI_tcbs[id]->intr=1; /* set interrupt */
		}
		/* kill is too costly to the root of the broadcast */
	}

	return 0;
}

/* signal the flag only when the flag is equal to oldval */
int signalAllConditionalFlag(int *pflag, int value, int oldval)
{
	register int status;
	int id=0, myid;

	if (*pflag!=oldval) return -1;

	if (!__compare_and_swap(pflag, oldval, value))
		return -1;

	myid=ukey_getid();

	for (; id<thrMPI_Nproc; id++) {
		if (id==myid) continue;

		status=thrMPI_tcbs[id]->status; 
		
		if (status==thrMPI_BUSY) {
			thrMPI_tcbs[id]->intr=1;
		}
		/* one by one kill is too costly */
	}

	return 0;
}

