/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT
#include <stdlib.h>
#include <sys/time.h>

# include "struct.h"
# include "mpi.h"

/* return sizeof a MPI type */
int thrMPI_Usize(MPI_Datatype unit) {	
  switch (unit) {
  case MPI_BYTE: return sizeof(char);
  case MPI_CHAR: return sizeof(char);
  case MPI_SHORT: return sizeof(short);
  case MPI_INT: return sizeof(int);
  case MPI_LONG: return sizeof(long);
  case MPI_UNSIGNED_LONG: return sizeof(unsigned long);
  case MPI_FLOAT: return sizeof(float);
  case MPI_DOUBLE: return sizeof(double);
  case MPI_LONG_DOUBLE: return sizeof(long double);

  /* deal with stupid FORTRAN */
  case MPI_DOUBLE_PRECISION: return sizeof(double);
  case MPI_INTEGER: return sizeof(int);
  case MPI_LOGICAL: return sizeof(int);
  default: 
# if (_DEBUG_)
	fprintf(stderr, "MPI_Usize: type not supported\n");
# endif
	return sizeof(int);
  }
}

/* Wall lock time is used through gettimeofday */
double thrMPI_Wtime() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return (double)t.tv_sec+t.tv_usec*0.000001;
}

double thrMPI_Wtick() {
  return 0.01;
}

int thrMPI_Init(int *argc, char ***argv) {
  return MPI_SUCCESS;
}

int thrMPI_Finalize() {
  return MPI_SUCCESS;
}

int MPI_Initialized(int *flag) {
  return MPI_SUCCESS;
}

int MPI_Abort(MPI_Comm comm, int errorcode) {
  return MPI_SUCCESS;
}

/* get the count of actual transmitted data */
int MPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int *count) {
  int size;
  size=thrMPI_Usize(datatype);
  if (size==0) {
    if (status->buflen > 0)
      (*count) = MPI_UNDEFINED;
    else
      /* This is ambiguous */
      (*count) = 0;
  }
  else {
    if ((status->buflen%(size)) != 0)
      (*count) = MPI_UNDEFINED;
    else
      (*count) = status->buflen/size;
  }
/**
	if (status) *count=status->count;
	else *count=MPI_UNDEFINED;
**/
  return MPI_SUCCESS;
}
