/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#ifndef MYTP_H
#define MYTP_H

#include <sys/types.h>
#include <sys/prctl.h>

#include "qt.h"

/* The notion of a thread is merged with the notion of a queue.
   Thread stuff: thread status (sp) and stuff to use during
   (re)initialization.  Queue stuff: next thread in the queue
   (next). */

/*typedef int mytp_lock_t;*/

typedef struct mytp_t {
  short thrid;
  short state;
  qt_t *sp;			/* QuickThreads handle. */
  void *sto;			/* `malloc'-allocated stack. */
  void *private;		/* thread-specific private data area */
  struct mytp_t *next;
#ifdef STAT
  int sync;
  int yield0;
  int yield1;
  int yield2;
  int yield3;
  int yield4up;
  int lastrun;
  int migrate;
#endif
} mytp_t;

/* data structure local to virtual processor */
typedef struct {
  int vp_id;
  mytp_t vp_main;	/* main thread for virtual processor */
  mytp_t *vp_curr;	/* currently-executing thread for virtual processor */
} mytp_vp_local_t;

#define mytp_vp_local ((mytp_vp_local_t*)(PRDA->usr_prda.fill))	/* virtual processor local data */
#define mytp_get_private() (mytp_vp_local->vp_curr->private)	/* get thread-specific data area */
#define mytp_get_myid() (mytp_vp_local->vp_curr->thrid)		/* get myid */

/* Each thread starts by calling a user-supplied function of this
   type. */
typedef void (mytp_userf_t)(void *p0);

/* Call this before any other primitives. */
extern void mytp_init();

/* When one or more threads are created by the main thread,
   the system goes multithread when this is called.  It is done
   (no more runable threads) when this returns. 
   The parameter specifies the number of kernel threads. */
extern void mytp_start (int);

/* Create a thread and make it runable.  When the thread starts
   running it will call `f' with arguments `p0' and `p0'. 
   The thread ID is returned. */
extern int mytp_create (mytp_userf_t *f, void *p0);

/* The current thread stops running but stays runable.
   It is an error to call `mytp_yield' before `mytp_start'
   is called or after `mytp_start' returns. */
extern void mytp_yield (void);

/* Like `mytp_yield' but the thread is discarded.  Any intermediate
   state is lost.  The thread can also terminate by simply
   returning. */
extern void mytp_abort (void);


/* Wall clock time */
extern double mytp_gettime (void);

#define IT_1us 33	/* 33 busy iterations takes about 1us */

extern void uspin (int);

/* initial a lock */
int mytp_lock_init (int*);

/* lock */
#define mytp_lock(a) while (__lock_test_and_set(a, 1) != 0) uspin(IT_1us)

/* locking, with a possibility of giving up the vp */
#define mytp_lock_block(a) while (__lock_test_and_set(a, 1) != 0) waitFlagEQ(a, 0, -1)

/* unlock */
#define mytp_unlock(a) __lock_release(a)

/* synchronization functions */
int waitFlagEQ (int*, int, int);
int waitFlagNE (int*, int, int);
int waitFlagIN (int*, int, int);
int waitFlagMOD (int*, int, int);

#define MAX_COND 100

typedef struct {
  int cur;
  int flag[MAX_COND];
} mytp_cond_t;

/* condition variable initialization */
int mytp_cond_init (mytp_cond_t*);

/* condition variable wait */
void mytp_cond_wait (mytp_cond_t*, int*);

/* condition variable broadcast */
void mytp_cond_bcast (mytp_cond_t*);

#endif /* ndef MYTP_H */
