/* mips.s -- assembly support. */

/*
 * QuickThreads -- Threads-building toolkit.
 * Copyright (c) 1993 by David Keppel
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/* Callee-save $16-$23, $30-$31.
 *
 * $25 is used as a procedure value pointer, used to discover constants
 * in a callee.  Thus, each caller here sets $25 before the call.
 *
 * On startup, restore regs so retpc === call to a function to start.
 * We're going to call a function ($4) from within this routine.
 * We're passing 3 args, therefore need to allocate 12 extra bytes on
 * the stack for a save area.  The start routine needs a like 16-byte
 * save area.  Must be doubleword aligned (_mips r3000 risc
 * architecture_, gerry kane, pg d-23).
 */

/*
 * Modified by Assar Westerlund <assar@sics.se> to support Irix 5.x
 * calling conventions for dynamically-linked code.
 */

	/* Make this position-independent code. */
	.option pic2

	.globl qt_block
	.globl qt_blocki
	.globl qt_abort
	.globl qt_start
	.globl qt_vstart

	/*
	** $4: ptr to function to call once curr is suspended
	**	and control is on $7's stack.
	** $5: 1'th arg to $4.
	** $6: 2'th arg to $4
	** $7: sp of thread to suspend.
	**
	** Totally gross hack: The MIPS calling convention reserves
	** 4 words on the stack for a0..a3.  This routine "ought" to
	** allocate space for callee-save registers plus 4 words for
	** the helper function, but instead we use the 4 words
	** provided by the function that called us (we don't need to
	** save our argument registers).  So what *appears* to be
	** allocating only 40 bytes is actually allocating 56, by
	** using the caller's 16 bytes.
	**
	** The helper routine returns a value that is passed on as the
	** return value from the blocking routine.  Since we don't
	** touch $2 between the helper's return and the end of
	** function, we get this behavior for free.
	*/
	.ent qt_blocki
qt_blocki:
	dsubu $sp, $sp,112	/* Allocate reg save space. */
	sd $16, 0+32($sp)
	sd $17, 8+32($sp)
	sd $18,16+32($sp)
	sd $19,24+32($sp)
	sd $20,32+32($sp)
	sd $21,40+32($sp)
	sd $22,48+32($sp)
	sd $23,56+32($sp)
	sd $30,64+32($sp)
	sd $31,72+32($sp)
qt_abort:
	move $2, $sp		/* $2 <= old sp to pass to func@$4. */
	move $sp, $7		/* $sp <= new sp. */
	.set noreorder
	move $25, $4		/* Set helper function procedure value. */
	jal $31,$25		/* Call helper func@$4 . */
	move $4, $2		/* $a0 <= pass old sp as a parameter. */
	.set reorder
	ld $31,72+32($sp)	/* Restore callee-save regs... */
	ld $30,64+32($sp)
	ld $23,56+32($sp)
	ld $22,48+32($sp)
	ld $21,40+32($sp)
	ld $20,32+32($sp)
	ld $19,24+32($sp)
	ld $18,16+32($sp)
	ld $17, 8+32($sp)
	ld $16, 0+32($sp)	/* Restore callee-save */

	daddu $sp,$sp,112	/* Deallocate reg save space. */
	j $31			/* Return to caller. */
	.end qt_blocki
	
	/*
	** Non-varargs thread startup.
	** Note: originally, 56 bytes were allocated on the stack.
	** The thread restore routine (_blocki/_abort) removed 40
	** of them, which means there is still 16 bytes for the
	** argument area required by the MIPS calling convention.
	*/
	.ent qt_start
qt_start:
	move $4, $16		/* Load up user function pu. */
	move $5, $17		/* ... user function pt. */
	move $6, $18		/* ... user function userf. */
	move $25, $19		/* Set `only' procedure value. */
	jal $31,$25		/* Call `only'. */
	dla $25,qt_error		/* Set `qt_error' procedure value. */
	j $25
	.end qt_start

	/*
	** Save calle-save floating-point regs $f24-$f31
	** See comment in `qt_block' about calling conventinos and
	** reserved space.  Use the same trick here, but here we
	** actually have to allocate all the bytes since we have to
	** leave 4 words leftover for `qt_blocki'.
	**
	** Return value from `qt_block' is the same as the return from
	** `qt_blocki'.  We get that for free since we don't touch $2
	** between the return from `qt_blocki' and the return from
	** `qt_block'.
	*/
	.ent qt_block
qt_block:
	dsubu $sp, $sp,80	/* 8 8-byte regs, saved ret pc, 16 bytes aligned. */
	s.d $f24,  0($sp)
	s.d $f25,  8($sp)
	s.d $f26, 16($sp)
	s.d $f27, 24($sp)
	s.d $f28, 32($sp)
	s.d $f29, 40($sp)
	s.d $f30, 48($sp)
	s.d $f31, 56($sp)
	sd $31, 64($sp)
	jal qt_blocki
	l.d $f24,  0($sp)
	l.d $f25,  8($sp)
	l.d $f26, 16($sp)
	l.d $f27, 24($sp)
	l.d $f28, 32($sp)
	l.d $f29, 40($sp)
	l.d $f30, 48($sp)
	l.d $f31, 56($sp)
	ld $31, 64($sp)
	daddu $sp, $sp,80
	j $31
	.end qt_block
