/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <strings.h>

#define SHM_KEY 111
#define SHM_SIZE 4096
#define MAX_APP 4

/* data structure for shmem area */

typedef struct {
  int total;
  int running;
  int recommended;
} app_t;

typedef struct {
  int daemonlock;
  int sys_total;
  int sys_running;
  app_t app[MAX_APP+1];
} daemon_t;

static void dump_shmem (void *shmvaddr) {
  int i;
  app_t *app_ptr;

  printf ("## shmem content: daemonlock=%d, sys_total=%d, sys_running=%d\n", ((daemon_t*)shmvaddr)->daemonlock, ((daemon_t*)shmvaddr)->sys_total, ((daemon_t*)shmvaddr)->sys_running);
  for (i=0; i<MAX_APP; i++) {
    app_ptr = ((daemon_t*)shmvaddr)->app + i;
    if (app_ptr->total != 0) 
      printf ("  APP %d: total=%d, running=%d, recommended=%d\n", i, app_ptr->total, app_ptr->running, app_ptr->recommended);
  }
}


main () {
  int shmid;
  void *shmvaddr;

  if ((shmid = shmget (SHM_KEY, SHM_SIZE, IPC_CREAT|SHM_R|SHM_W|044)) == -1) {
    perror ("shmget");
    exit (1);
  }
	
  if ((shmvaddr = shmat (shmid, (void *)0, 0)) == (void *)(-1)) {
    perror ("shmat");
    exit (1);
  }

  dump_shmem (shmvaddr);

  bzero (shmvaddr, SHM_SIZE);

  dump_shmem (shmvaddr);

  if (shmdt (shmvaddr) == -1) {
    perror ("shmdt");
    exit (1);
  }
}
