/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#include <sys/time.h>

#include "mytp.h"

/* Wall clock time */
double mytp_gettime () {
  struct timeval t;
  gettimeofday (&t, NULL);
  return (double) t.tv_sec+t.tv_usec*0.000001;
}

/* initial a lock */
int mytp_lock_init (int *var) {
  *var=0;
  return 0;
}

/* lock */
void mytp_lock (int *a) {
  while (__lock_test_and_set(a, 1) != 0) uspin(IT_1us);
}

/* condition variable initialization */
int mytp_cond_init (mytp_cond_t *cond) {
  cond->cur=0;
  cond->flag[0]=0;
  return 0;
}

/* condition variable wait */
void mytp_cond_wait (mytp_cond_t *cond, int *lock) {
  register int *pos;

  pos = cond->flag+cond->cur;
  while (*pos == 0) {
    mytp_unlock (lock);
    waitFlagEQ (pos, 1, -1);
    mytp_lock_block (lock);
  }
}

/* condition variable broadcast */
void mytp_cond_bcast (mytp_cond_t *cond) {
  cond->flag[cond->cur] = 1;
  cond->cur = (cond->cur+1)%MAX_COND;
  cond->flag[cond->cur]=0;
}
