/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _UTIL_H_
# define _UTIL_H_

#ifdef PTHREAD 
#include <pthread.h>
#endif

#ifdef MYTP
#include "mytp.h"
#define pthread_mutex_t int
#define pthread_mutex_init(a,b) mytp_lock_init(a)
#define pthread_mutex_lock(a) mytp_lock_block(a)
#define pthread_mutex_unlock(a) mytp_unlock(a)
#define pthread_mutex_destroy(a) 0
#define pthread_cond_t mytp_cond_t 
#define pthread_cond_init(a,b) mytp_cond_init(a)
#define pthread_cond_wait(a,b) mytp_cond_wait(a,b)
#define pthread_cond_broadcast(a) mytp_cond_bcast(a)
#define pthread_cond_destroy(a) 0
#endif

typedef char BOOL;
enum {FALSE=0, TRUE};

# define MAX(a,b) (((a)>=(b))?(a):(b))
# define MIN(a,b) (((a)<=(b))?(a):(b))

#define WAIT(MUTEX)   pthread_mutex_lock(&(MUTEX))
#define SIGNAL(MUTEX)     pthread_mutex_unlock(&(MUTEX))
typedef pthread_mutex_t MUTEX;
typedef pthread_cond_t COND;

typedef struct {
	int	maxcnt;                 /* maximum number of runners    */
	struct	_sb {
	COND	wait_cv;        /* cv for waiters at barrier    */
	MUTEX	wait_lk;        /* mutex for waiters at barrier */
	int	runners;        /* number of running threads    */
	} sb[2];
	struct _sb	*sbp;           /* current sub-barrier          */
} BARRIER;
	
/* a mailbox that can only hold one message */
typedef struct {
	int from; 
	char *data;
	int size;
	MUTEX mutex;
	MUTEX empty;
	MUTEX full;
	MUTEX fnsh;
} MBOX;
	
/* broadcast control object */
typedef struct {
	int avail; /* whether the data is available, >0 available, =0 not available */
	COND	cond;
	MUTEX mutex;
	char *data;
	/** int size; **/
	BARRIER bar;
} BCOBJ;

int bar_init(BARRIER *bp, int count);
int bar_wait(BARRIER *bp);
int bar_destroy( BARRIER *bp );

int mbox_init(MBOX *mbox);
int mbox_destroy(MBOX *mbox);
int mbox_recv(MBOX *mbox, char *data, int size, int from);
int mbox_arecv(MBOX *mbox, char *data, int size, int from);
int mbox_send(MBOX *mbox, char *data, int size, int from);
int mbox_asend(MBOX *mbox, char *data, int size, int from);

int bco_init(BCOBJ *bco, int maxcnt);
int bco_destroy(BCOBJ *bco);
int bco_bcast(BCOBJ *bco, char *data, int size, int snder, int me);
int bco_aggbcast(BCOBJ *bco, char *data, int size, int snder, int me);

double get_wtime();
# endif
