/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#ifdef PTHREAD
#include <pthread.h>
#endif

#ifdef MYTP
#include <mytp.h>
#endif

typedef struct {
  int id;
  int noproc;
} parm;

double gettime() {
  struct timeval t;
  gettimeofday (&t, NULL);
  return (double) t.tv_sec+t.tv_usec*0.000001;
}

double *finals;
int rootn;

double f (double a) {

/* *((int*)mytp_get_private()) = *((int*)mytp_get_private())+1; */

  return (4.0 / (1.0 + a * a));
}

void cpi (void *arg) {
  parm      *p = (parm *) arg;
  int        myid = p->id;
  int        numprocs = p->noproc;
  int        i;
  double     h, sum, x;

/* *((int*)mytp_get_private())=0; */

  h = 1.0 / (double) rootn;
  sum = 0.0;
  for (i = myid + 1; i <=rootn; i += numprocs) {
    x = h * ((double) i - 0.5);
    sum += f(x);
  }
  finals[myid] = h*sum;

/* printf ("myid=%d, f is called %d times\n", myid, *((int*)mytp_get_private())); */
}

void main (int argc, char *argv[]) {
  int n, nproc, i;
  parm *arg;
#ifdef PTHREAD
  pthread_t *threads;
#endif
  double startwtime, endwtime;
  double PI25DT = 3.141592653589793238462643;
  double pi;

  if (argc==4) {
    n=atoi(argv[1]);
    nproc=atoi(argv[2]);
    rootn=atoi(argv[3]);
  }
  else {
    printf("usage: %s <#threads> <#nproc> <#iterations>\n", argv[0]);
    exit(-1);
  }

  printf("\t#threads=%d\n", n);

  /* allocate spaces */
#ifdef PTHREAD
  threads = (pthread_t *) malloc(n * sizeof(*threads));
#endif
  finals = (double *) malloc(n * sizeof(double));
  arg = (parm *) malloc(sizeof(parm)*n);

#ifdef PTHREAD
  /* set concurrency level */
  pthread_setconcurrency(n);
#endif

  startwtime = gettime();

  /* Spawn thread */
#ifdef MYTP
  mytp_init ();
#endif
  for (i = 0; i < n; i++) {
    arg[i].id = i;
    arg[i].noproc = n;
#ifdef MYTP
    mytp_create (cpi, (void *)(arg+i));
#endif
#ifdef PTHREAD
    pthread_create(&threads[i], NULL, cpi, (void *)(arg+i));
#endif
  }
#ifdef MYTP
  mytp_start (nproc);
#endif

#ifdef PTHREAD
  /* Synchronize the completion of each thread. */
  for (i = 0; i < n; i++) {
    pthread_join (threads[i], NULL);
  }
#endif

  pi = 0.0;
  for (i =0; i < n; i++)
    pi += finals[i];
  endwtime = gettime();
  printf("\tPI=%.16e, ERR=%.16e\n", pi, fabs(pi - PI25DT));
  printf("\tTIME=%e\n", (endwtime - startwtime));

  free (finals);
  free (arg);
}

